/*
 * Copyright (c) 2008  Yuri Chernyavsky
 * Copyright (c) 2007  Gregor Richards
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>
#include <dirent.h>
#include <libgen.h>
#include <stdarg.h>
#include <assert.h>
#include <locale.h>

#include "smv_encode.h"
#include "i18n.h"

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

typedef struct {
  int w;
  int h;
  int fr;
  int fpj;
  int q;
  int rotate;
} params_s; 

typedef FILE* (*parser_t) (const char*);

static int dflag = 0;
static int bflag = 0;
char *targ = NULL;

/* Test if the input video file exists */
static int file_exists(const char *name) {
  struct stat s_buf;
  return (stat(name, &s_buf) != -1);
}

/* Test if the output file is writable */
static int file_writable(const char *name) {
  FILE *f = fopen(name, "w");
  if (f) {
     fclose(f);
     unlink(name);
     return 1;
  }
  return 0;
}

/* Messages */

const char *phasestr(phase_t p) {
  if (p == phase_decode)
    return i18n("Decode input video");
  if (p == phase_frames)
    return i18n("Frames");
  if (p == phase_tile)
    return i18n("Tile frames");
  if (p == phase_audio)
    return i18n("Generate audio file");
  if (p == phase_combine)
    return i18n("Generate SMV file");
  return "";
}

const char *errstr(err_t e) {
  if (e == err_input_file)
    return i18n("Input video not found\n");
  if (e == err_output_file_not_writable) 
    return i18n("Output file cannot be opened for writing\n");
  if (e == err_bad_parameter)
    return i18n("Invalid parameters\n");
  if (e == err_decode || e == err_frames || e == err_tile || e == err_audio ||
      e == err_combine)
    return i18n("Combine failed\n");
  if (e == err_unexpected)
    return i18n("Unexpected error\n");
  return "";
}

/* Send event to gui frontend */
static void event(tag_t tag, int type) {
  msg_t msg = { (msg_tag_t)tag, (msg_content_t)type };

  if (bflag) {
    /* Send message to the gui via stdout */
    fwrite(&msg, sizeof(msg_t), 1, stdout);
    fflush(stdout);
    return;
  }

  switch(tag) {
  case error:
    fprintf(stdout, "%s", errstr((err_t)type));
    return;
  case interrupted:
    fprintf(stdout, i18n("Encoder interrupted\n"));
    return;
  case success:
    fprintf(stdout, i18n("Finished successfully\n"));
    break;
  }

  if (!dflag) {
     switch(tag) {
     case phase_start:
     case phase_start_percentage:
       fprintf(stdout, "%s[", phasestr((phase_t)type));  
       break;
     case phase_end:
       fprintf(stdout, "]\n");  
       break;
      
     case alive:
     case percentage:
       fprintf(stdout, ".");  
     default:
       break; 
     }
  }
  fflush(stdout);
}

/* Various events */
static void err(err_t type) {
  event(error, (msg_content_t)type);
  exit(EXIT_STATUS_ERROR);
}

static void event_phase_start(phase_t type, int percentage) {
  if (percentage) {
    event(phase_start, (msg_content_t)type);
  } else {
    event(phase_start_percentage, (msg_content_t)type);
  }
}

static void event_phase_end(phase_t type) {
  event(phase_end, (msg_content_t)type);
}

static void event_alive() {
  event(alive, 0);
}

static void event_percentage(int pc) {
  event(percentage, pc);
}


static void dump(const char *msg) {
  if (dflag) {
     fprintf(stdout, "DEBUG %s\n", msg);
     fflush(stdout);	
  }
}

/* Temporary folder for intermediate files */
#ifndef PATH_MAX
#define PATH_MAX 4096
#endif

static char pwd[PATH_MAX];
static char *tmp;
static void init() {
  /* Remember the current working directory to create the output smv file there */
  if (!getcwd(pwd, PATH_MAX)) err(err_unexpected);

  if (!targ) {
      tmp = strdup("/tmp/smv_encode.XXXXXX");
      if (!mkdtemp(tmp)) err(err_unexpected);
  } else {
      tmp = targ;
      mkdir(tmp, S_IRWXU);
      if (!file_exists(tmp)) err(err_unexpected);
  }
}

/* Normal system() blocks SIGINT and SIGQUIT */
int mysystem(const char *cmd) {
    int ret = system(cmd);

    if (WIFEXITED(ret)) {
        return WEXITSTATUS(ret); 
    } 

    if (WIFSIGNALED(ret) &&
        (WTERMSIG(ret) == SIGINT || WTERMSIG(ret) == SIGQUIT)) {   
        raise(WTERMSIG(ret));
    }
    return -1;
}

static void cleanup() {
  if (tmp && !targ) {
        /* No -t flag, the temp dir name is short */
        char rm_cmd[100];
        sprintf(rm_cmd, "rm -rf %s", tmp);
        system(rm_cmd); 
  }
  tmp = NULL;
}

/* Maybe too heavy for signal handler - but works */
static void handle_int(int sig) {
  event(interrupted, 0);
  exit(EXIT_STATUS_INTERRUPTED);
}

/* Filter for directory scanning using scandir() */
static int file_select(const struct dirent *entry) {
  if ((strcmp(entry->d_name, ".") == 0) || 
      (strcmp(entry->d_name, "..") == 0))
    return 0;
  else
    return 1;
}

/* Print int as three bytes in little-endian order */
static void print_int(int v, FILE *out)
{
  fprintf(out, "%c", v % 256);
  v >>= 8;
  fprintf(out, "%c", v % 256);
  v >>= 8;
  fprintf(out, "%c", v % 256);
}

/* Allocate memory and create a formatted string
   taken from vsnprintf man page */
char *make_str(const char *fmt, ...) {
  /* Guess we need no more than 100 bytes. */
  int n, size = 1000;
  char *p, *np;
  va_list ap;

  if ((p = malloc(size)) == NULL)
    return NULL;

  while (1) {
    /* Try to print in the allocated space. */
    va_start(ap, fmt);
    n = vsnprintf(p, size, fmt, ap);
    va_end(ap);
    /* If that worked, return the string. */
    if (n > -1 && n < size)
      return p;
    /* Else try again with more space. */
    if (n > -1)    /* glibc 2.1 */
      size = n+1; /* precisely what is needed */
    else           /* glibc 2.0 */
      size *= 2;  /* twice the old size */
    if ((np = realloc (p, size)) == NULL) {
      free(p);
      return NULL;
    } else {
      p = np;
    }
  }
}


/* Command line for the external software we use */
char* get_decode_cmdline(const params_s* params, const char* inp, const char *workdir, 
                         const char* video) {
  char rotatestr[100];
  rotatestr[0] = '\0'; 

  if (params -> rotate == -2) {
     sprintf(rotatestr, "flip,");
  } else if (params -> rotate != -1) {
     sprintf(rotatestr, "rotate=%d,", params -> rotate);
  }

  char *cmd = make_str(
"mencoder \"%s\"  -ovc lavc -oac pcm -lavcopts vcodec=ffv1 -ofps %d -vf %sexpand=:::::%d/%d,scale=%d:%d -sws 9  -af resample=11025 -o %s/%s", 
          inp, params->fr, rotatestr, params->w, params->h, params->w, params->h, 
          workdir, video);
  assert(cmd);
  return cmd;
}

char* get_frames_cmdline(const params_s* params, const char* video, const char *workdir, 
                         const char* outdir) {
  char *cmd = make_str(
"mplayer %s/%s  -slave -vo jpeg:quality=%d::outdir=\"%s/%s\":maxfiles=1000000 -nosound -ac null -ao null -benchmark", 
          workdir, video, params->q, workdir, outdir);
  assert(cmd);
  return cmd;
}

char* get_tile_cmdline(const params_s* params, const char* inp, const char *workdir, 
                       const char* video) {
  char *cmd = make_str(
"mencoder %s/%s -ovc lavc -oac copy -lavcopts vcodec=ffv1 -vf tile=1:%d:%d:0:0 -o %s/%s",
          workdir, inp, params->fpj, params->fpj, workdir, video);
  assert(cmd);
  return cmd;
}

char* get_audio_cmdline(const params_s* params, const char* video, const char *workdir, 
                        const char* audio) {
  char *cmd = make_str(
"ffmpeg -y -i %s/%s -vn -acodec adpcm_ima_wav -ab 89416 %s/%s", workdir, video, workdir, audio);
  assert(cmd);
  return cmd;
}

/* Reap the process started with popen() */
int pclose_status(FILE *fp) {
  int status;

  if ((status = pclose(fp)) < 0) err(err_unexpected);
  if (WIFEXITED(status)) {
    return WEXITSTATUS(status);
  }

  if (WIFSIGNALED(status)) {
    return -1;
  }
  return -1;
}

#define BUF_SIZE 1024
FILE* alive_parser(const char* cmd) {
  FILE *fp;
  char buf[BUF_SIZE + 1];
  char *the_cmd = make_str("%s 2>/dev/null", cmd);

  if ((fp = popen(the_cmd, "r")) == NULL) err(err_unexpected);
  free(the_cmd);

  while (fgets(buf, BUF_SIZE, fp) != NULL) event_alive();
  return fp;
}

/* We assume that the percentage comes in the following format: \d?\d?\d%) 
   An alternative is to use awk */
FILE* mencoder_percentage_parser(const char *cmd) {
    FILE *fp;
    char pbuf[4] = {'0','0','0'};
    char ch;
    int percent = 0, newpercent = 0;
    char *the_cmd = make_str("%s 2>/dev/null", cmd);

    if ((fp = popen(the_cmd, "r")) == NULL) err(err_unexpected);
    free(the_cmd);

    while ((ch=fgetc(fp)) != EOF) {
      if (ch == '%') {
        ch = fgetc(fp);
        if (ch == EOF) break;
        if (ch != ')') {
            ungetc(ch, fp);
            continue;
        }
        newpercent = atoi(pbuf);
        if (newpercent > percent && newpercent <= 100) {
           event_percentage(newpercent);
           percent = newpercent;
        }
      }
      else if (ch >= '0' && ch <= '9') {
        pbuf[0] = pbuf[1]; pbuf[1] = pbuf[2]; pbuf[2] = ch;
      }
      else {
        sprintf(pbuf, "%c%c%c", '0', '0', '0');
      }
    }
    
    return fp;
}

/* Execute an external command and parse its output */
static int do_externalcmd_phase(phase_t phase_name, parser_t parser, const char* cmd) {
  int ret;
  
  if (dflag) {
    dump(cmd);
    return mysystem(cmd);
  } 
  
  event_phase_start(phase_name, (parser == alive_parser));
  if (!(ret = pclose_status(parser(cmd)))) {
    event_phase_end(phase_name);
  }
  return ret;
}

static void decode_input_video(const params_s* params,
		            const char* inp, const char* output_dir) {
  char* cmd;
  int ret;
  const char *videostr = "video.avi";
  const char *tvideostr = "video_tiled.avi";
  const char *audiostr = "audio.wav";
  const char *framesstr = "frames";

  /* Cleanup temp directory */
  if (!chdir(tmp)) {
      system("rm -rf video.avi video_tiled.avi audio.wav frames");
      chdir(pwd);
  }

  /* call mencoder */
  cmd = get_decode_cmdline(params, inp, tmp, videostr); 
  ret = do_externalcmd_phase(phase_decode, mencoder_percentage_parser, cmd);
  free(cmd);
  if (ret) err(err_decode);

  /* tile */
  if (params->fpj > 1) {
     cmd = get_tile_cmdline(params, videostr, tmp, tvideostr);
     ret = do_externalcmd_phase(phase_tile, mencoder_percentage_parser, cmd);
     free(cmd);
     if (ret) err(err_tile);
  } else {
     tvideostr = videostr;
  }

  /* call mplayer */
  cmd = get_frames_cmdline(params, tvideostr, tmp, framesstr);
  ret = do_externalcmd_phase(phase_frames, alive_parser, cmd);
  free(cmd);
  if (ret) err(err_frames);
  
  /* call ffmpeg */
  cmd = get_audio_cmdline(params, videostr, tmp, audiostr);
  ret = do_externalcmd_phase(phase_audio, alive_parser, cmd);
  free(cmd);
  if (ret) err(err_audio); 
}

/* Create a smv file from audio file and jpeg frames */
/* Gregor Richards' code */
static void encode_smv(const params_s* params, const char* input_dir, const char* smv_name) {
  FILE *fhnd, *smv;
  int c, i;
  int fc, js = 0;
  struct dirent **files;
  off_t *sizes;
  struct stat s_buf;

  event_phase_start(phase_combine, 0);
  if (chdir(pwd) != 0) err(err_unexpected);
  smv = fopen(smv_name, "wb");

  if  (chdir(input_dir) != 0) err(err_unexpected);
  /* first just cat in the .wav file */
  fhnd = fopen("audio.wav", "r");
  if (fhnd != NULL) {
      while ((c = fgetc(fhnd)) != EOF) {
        fputc(c, smv);
      }
      fclose(fhnd);
  }

  /* check the sizes of the jpeg files */
  if (chdir("frames") != 0) err(err_unexpected);

  fc = scandir(".", &files, file_select, alphasort);
  if (fc == 0) err(err_unexpected);

  sizes = (off_t*)malloc(sizeof(off_t) * fc);
  for (i=0; i<fc; i++) {
    if (stat(files[i]->d_name, &s_buf) == -1) err(err_unexpected);
    sizes[i] = s_buf.st_size;
    if (sizes[i] > js) {
      js = sizes[i];
    }
  }
  js ++;
  
  /* basic header */
  fwrite("SMV002000", 1, 9, smv);
  
  /* width */
  print_int(params->w, smv);
  
  /* height */
  print_int(params->h, smv);
  
  /* unidentified */
  fwrite("\x10\x00\x00\x01\x00\x00", 1, 6, smv);
  
  /* jpeg size */
  print_int(js+3, smv);
  
  /* frame rate */
  print_int(params->fr, smv);
  
  /* # frames */
  print_int(fc * params->fpj, smv);
  
  /* unidentified */
  fwrite("\x01\x00\x00\x00\x00\x00", 1, 6, smv);
  
  /* frames per jpeg */
  print_int(params->fpj, smv);
  
  /* unidentified */
  fwrite("\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01\x01", 1, 12, smv);
  
  /* the jpeg files */
  for (i=0; i<fc; i++) {
    
    /* write size */
    print_int(sizes[i], smv);
    fhnd = fopen(files[i]->d_name, "r");
    if (fhnd == NULL) err(err_unexpected);

    /* write jpeg */
    while ((c = fgetc(fhnd)) != EOF) {
      fputc(c, smv);
    }
    fclose(fhnd);
	
    /* Write padding. We use NULL bytes, the proprietary
       encoders does not, however there is no difference in
       the behavior. */
    for (c = sizes[i]; c < js; c++) {
      fputc(0, smv);
    }
  }

  fclose(smv);
  event_phase_end(phase_combine);
}

/* Print version */
static void version() {
  fprintf(stdout, "%s\n", VERSION);
  exit(EXIT_STATUS_NO_RUN);
}

/* Print usage */
static void usage(const char *app_name, int full) {
  if (!full) {
    fprintf(stdout, i18n(
"SigmaTel Motion Video encoder (%s)\n\
Usage: %s [OPTIONS]... input_file [output_file]\n\n\
To produce movie.smv from movie.avi:\n\
       %s movie.avi\n\n\
To print full list of options:\n\
       %s -l\n"),
	    VERSION, app_name, app_name, app_name);
  } else {
    fprintf(stdout, i18n( 
"SigmaTel Motion Video encoder (%s)\n\
Usage: %s [OPTIONS]... input_file [output_file]\n\n\
Options: (use defaults unless you know what you are doing)\n\
   -v print version and exit\n\
   -l print this screen and exit\n\
   -g <geometry>: default %dx%d\n\
   -f <framerate>: default %d\n\
   -n <frames per jpeg file>: defaults to the framerate (if specified, otherwise %d)\n\
   -q <quality>: 0-100, recommended <=85, default %d\n\
   -r <rotate>: default %d\n\
      -1    No rotation\n\
      -2    Flip upside-down.\n\
       1    Rotate by 90 degrees clockwise.\n\
       2    Rotate by 90 degrees counterclockwise.\n\
\n\
For debugging:\n\
   -d (dump) print output of mencoder, mplayer, ffmpeg\n\
   -t <temp_dir>: use <temp_dir> as the temporary folder; do not remove it\n\n\
   -b (backend mode) run with a gui frontend\n"),
     VERSION, app_name, DEFAULT_PARAM_W, DEFAULT_PARAM_H, DEFAULT_PARAM_FR,
     DEFAULT_PARAM_FPJ, DEFAULT_PARAM_Q, DEFAULT_PARAM_ROTATE);
  }
  exit(EXIT_STATUS_NO_RUN);
}

/* Main */
int main (int argc, char **argv) {
  int c; 
  extern char *optarg; 
  extern int optind; 
 
  /* initialize gettext */
  setlocale (LC_ALL, "");
  bindtextdomain (PACKAGE, LOCALEDIR);
  textdomain (PACKAGE);
 
  int errflg = 0; 
  params_s params = {DEFAULT_PARAM_W, 
                     DEFAULT_PARAM_H, 
                     DEFAULT_PARAM_FR,
                     DEFAULT_PARAM_FPJ, 
                     DEFAULT_PARAM_Q, 
                     DEFAULT_PARAM_ROTATE}; 
  char *inp, *out;

#ifdef DISABLE_UI
  dflag = 1;
#endif

  if (argc == 1) {
      usage(argv[0], 0);
  }

  while ((c = getopt(argc, argv, "lt:vbdg:f:n:q:r:")) != EOF) {
    switch (c) {
    case 'v': version(); break;
    case 'l': usage(argv[0], 1); break;
    case 'b': bflag = 1; break;
    case 'd': dflag = 1; break;
    case 't': targ = strdup(optarg); break;
    case 'g': 
	{
            char *arg, *w_str, *h_str;

	    w_str = arg = strdup(optarg);
            params.h = 0;
            h_str = strchr(arg, 'x');
            if (h_str) {
               *h_str = '\0';
               h_str++;
               params.h = atoi(h_str); 
            }
            params.w = atoi(w_str); 
            free(arg); 
        }
        break;
    case 'f': params.fr = atoi(optarg); break;
    case 'n': params.fpj = atoi(optarg); break;
    case 'q': params.q = atoi(optarg); break;
    case 'r': params.rotate = atoi(optarg); break;
    case '?': errflg++; 
    }
  }

  event(hello, 0);

  if (errflg || optind == argc || optind < argc - 2) { 
    err(err_bad_parameter);
  }  

  if (!params.w || !params.h || !params.fr || !params.fpj) {
    err(err_bad_parameter);
  }

  inp = strdup(argv[optind]);
  /* If output filename is specified */
  if (optind == argc - 2) {
    out = strdup(argv[optind + 1]);
  } else {
    char *str = basename(strdup(inp));
    out = (char*)malloc(strlen(str) + 5);
    sprintf(out, "%s.smv", str);
  }

  /* Check if input file exists */
  if (!file_exists(inp)) err(err_input_file);

  /* Check if output file writable */
  if (!file_writable(out)) err(err_output_file_not_writable);

  /* Temporary folder - always remove it on exit */
  init();
  atexit(cleanup);
  /* Gui may cancel the encoding by sending us SIGINT signal
     we have to cleanup temp dir in this case */
  signal(SIGINT, handle_int);

  /* Extract audio and video frames */
  decode_input_video(&params, inp, tmp);
  /* Create smv video */
  encode_smv(&params, tmp, out);

  event(success, 0);
  return 0;
}



